//each parameter has a "ui value" that the user views and interacts with that has a smaller range than the "real value" that the camera module actually uses
struct ParamData
{
  byte digits; //number of digits to be displayed: signed 1 digit value, or unsigned 2 digits value
  const uint8_t* bitmap; //icon used as legend for this parameter
  int16_t ui_value = 0, min_ui, max_ui, min_real, max_real;
  const char* key; //string for identifying this parameter in ESP32 preferences storage
  ParamData() {}
  ParamData(byte _digits, const uint8_t* _bitmap, int16_t _min_ui, int16_t _max_ui, int16_t _min_real, int16_t _max_real, const char* _key)
  {
    digits = _digits;
    bitmap = _bitmap;
    min_ui = _min_ui;
    max_ui = _max_ui;
    min_real = _min_real;
    max_real = _max_real;
    key = _key;
  }
  ParamData(byte _digits, const uint8_t* _bitmap, int16_t _min, int16_t _max, const char* _key)
  {
    digits = _digits;
    bitmap = _bitmap;
    min_ui = min_real = _min;
    max_ui = max_real = _max;
    key = _key;
  }
};


#define NUM_PARAMS 8
#define PARAM_FOCUS 0
#define PARAM_EV 1
#define PARAM_SHUTTER 2
#define PARAM_GAIN 3
#define PARAM_WHITEBALANCE 4
#define PARAM_BRIGHTNESS 5
#define PARAM_CONTRAST 6
#define PARAM_SATURATION 7

ParamData params[NUM_PARAMS];
void setupParams()
{
  params[PARAM_FOCUS] =        ParamData(2, bmp_focus, 0, 50, 255, 0, NULL);
  params[PARAM_EV] =           ParamData(1, bmp_ev, -5, 5, "ev");
  params[PARAM_SHUTTER] =      ParamData(2, bmp_shutter, 0, 50, 0, 984, "shutter");
  params[PARAM_GAIN] =         ParamData(2, bmp_gain, 0, 50, 0, 64, "gain");
  params[PARAM_WHITEBALANCE] = ParamData(2, bmp_wb, 1, 4, "wb");
  params[PARAM_BRIGHTNESS] =   ParamData(1, bmp_brightness, -3, 3, "brightness");
  params[PARAM_CONTRAST] =     ParamData(1, bmp_contrast, -3, 3, "contrast");
  params[PARAM_SATURATION] =   ParamData(1, bmp_saturation, -4, 4, "saturation");
}


void setUIValue(byte p, int16_t new_value)
{
  params[p].ui_value = constrain(new_value, params[p].min_ui, params[p].max_ui);
  setRealValue(p, map(params[p].ui_value, params[p].min_ui, params[p].max_ui, params[p].min_real, params[p].max_real));
}


void incrementUIValue(byte p, int16_t change)
{
  setUIValue(p, params[p].ui_value + change);
}


int16_t getUIValue(byte p)
{
  params[p].ui_value = map(getRealValue(p), params[p].min_real, params[p].max_real, params[p].min_ui, params[p].max_ui);
  return params[p].ui_value;
}


void setRealValue(byte p, int16_t new_value)
{
  switch(p)
  {
    case PARAM_FOCUS: if(!isInAuto(p)) { setManualFocus(new_value); } break;
    case PARAM_EV: camera->set_ae_level(camera, new_value); break;
    case PARAM_SHUTTER:
      if(isInAuto(p)) { camera->set_exposure_ctrl(camera, 0); }
      camera->set_aec_value(camera, new_value);
      break;
    case PARAM_GAIN:
      if(isInAuto(p)) { camera->set_gain_ctrl(camera, 0); }
      camera->set_agc_gain(camera, new_value);
      break;
    case PARAM_WHITEBALANCE: camera->set_wb_mode(camera, new_value); break;
    case PARAM_BRIGHTNESS: camera->set_brightness(camera, new_value); break;
    case PARAM_CONTRAST: camera->set_contrast(camera, new_value); break;
    case PARAM_SATURATION: camera->set_saturation(camera, new_value); break;
  }
}


int16_t getRealValue(byte p)
{
  switch(p)
  {
    case PARAM_FOCUS: return getFocusPosition(); break;
    case PARAM_EV: return camera->status.ae_level; break;
    case PARAM_SHUTTER: return camera->status.aec_value; break;
    case PARAM_GAIN: return camera->status.agc_gain; break;
    case PARAM_WHITEBALANCE: return camera->status.wb_mode; break;
    case PARAM_BRIGHTNESS: return camera->status.brightness; break;
    case PARAM_CONTRAST: return camera->status.contrast; break;
    case PARAM_SATURATION: return camera->status.saturation; break;
  }
}


//check if parameter is in auto mode (true) or manual mode (false)
byte isInAuto(byte p)
{
  switch(p)
  {
    case PARAM_FOCUS: return getAutofocusStatus() == 0; break;
    case PARAM_SHUTTER: return camera->status.aec; break;
    case PARAM_GAIN: return camera->status.agc; break;
    case PARAM_WHITEBALANCE: return getRealValue(p) == 0; break;
    default: return false;
  }
}


//set parameter to auto mode, or reset it to back to zero if it doesn't have an auto mode
void reset(byte p)
{
  switch(p)
  {
    case PARAM_FOCUS: triggerAutofocus(); break;
    case PARAM_SHUTTER: camera->set_exposure_ctrl(camera, 1); break;
    case PARAM_GAIN: camera->set_gain_ctrl(camera, 1); break;
    default: setRealValue(p, 0);
  }
  getUIValue(p);
}

#define IN_AUTO -32768 //save or load this value to indicate that the saved parameter is in auto mode
//load parameters from ESP32 preferences storage
void loadParams()
{
  preferences.begin("blossom", true);
  for(byte i = 0; i < NUM_PARAMS; i++)
  {
    if(params[i].key)
    {
      int16_t value = preferences.getShort(params[i].key, IN_AUTO);
      if(value == IN_AUTO) { reset(i); }
      else { setUIValue(i, value); }
    }
  }
  preferences.end();
}

//save parameters to ESP32 preferences storage. these are saved after a short delay after a parameter has been modified, or whenever a photo has been taken
void saveParams()
{
  param_save_time = 0; //parameters are now being saved, so disarm parameter save timer
  preferences.begin("blossom", false);
  for(byte i = 0; i < NUM_PARAMS; i++)
  {
    if(params[i].key)
    {
      preferences.putShort(params[i].key, isInAuto(i) ? IN_AUTO : params[i].ui_value);
    }
  }
  preferences.end();
}

void armParamSaveTimer(byte p)
{
  if(params[p].key) { param_save_time = millis() + PARAM_SAVE_DELAY; } 
}