class Option
{
  public:
  const char* label; //text displayed for this option in the menu
  const char* key; //string for identifying this parameter in ESP32 preferences storage
  bool enabled;
  Option(const char* _label, const char* _key, bool default_value)
  {
    label = _label;
    key = _key;
    enabled = default_value;
  }
};

#define NUM_OPTIONS 4
Option* option_shutter_af = new Option("shutter af", "shutteraf", 0); //perform autofocus whenever shutter is pressed, before taking a photo
Option* option_thirds_grid = new Option("thirds grid", "thirdsgrid", 0); //show thirds grid
Option* option_horizon = new Option("horizon", "horizon", 0); //show accelerometer horizon level
Option* option_beep = new Option("beep", "beep", 1); //show accelerometer horizon level
Option* options[] = { option_shutter_af, option_thirds_grid, option_horizon, option_beep };

//load options from ESP32 preferences storage
void loadOptions()
{
  preferences.begin("blossom", true);
  for(byte i = 0; i < NUM_OPTIONS; i++)
  {
    if(options[i]->key)
    {
      options[i]->enabled = preferences.getBool(options[i]->key, options[i]->enabled);
    }
  }
  preferences.end();
}

//save options to ESP32 preferences storage. these are saved when the options menu is closed
void saveOptions()
{
  preferences.begin("blossom", false);
  for(byte i = 0; i < NUM_OPTIONS; i++)
  {
    if(options[i]->key)
    {
      preferences.putBool(options[i]->key, options[i]->enabled);
    }
  }
  preferences.end();
}

#define OPTIONSMENU_PADDING 12L //space around outer edge of menu
#define OPTIONSMENU_MARGIN 10 //space between border and inner text
#define OPTIONSMENU_RADIUS 10 //corner radius
#define OPTIONSMENU_LINEHEIGHT 10 //character height
#define OPTIONSMENU_LINESPACE 5 //space between lines of text
#define OPTIONSMENU_WIDTH 120
#define OPTIONSMENU_HEIGHT (OPTIONSMENU_MARGIN * 2 + OPTIONSMENU_LINEHEIGHT * NUM_OPTIONS + OPTIONSMENU_LINESPACE * (NUM_OPTIONS - 1))
byte optionsmenu_open = 0;
int8_t optionsmenu_cursor = 0;
uint32_t optionsmenu_open_time = 0;

void drawOptionsMenu()
{
  canvas.fillRoundRect(OPTIONSMENU_PADDING, OPTIONSMENU_PADDING, OPTIONSMENU_WIDTH, OPTIONSMENU_HEIGHT, OPTIONSMENU_RADIUS, ST77XX_BLACK);
  canvas.drawRoundRect(OPTIONSMENU_PADDING + 1, OPTIONSMENU_PADDING + 1, OPTIONSMENU_WIDTH - 2, OPTIONSMENU_HEIGHT - 2, OPTIONSMENU_RADIUS - 2, COLOR_HIGHLIGHT);
  byte y = OPTIONSMENU_PADDING + OPTIONSMENU_MARGIN + OPTIONSMENU_LINEHEIGHT;
  for(byte i = 0; i < NUM_OPTIONS; i++)
  {
    canvas.setCursor(OPTIONSMENU_PADDING + OPTIONSMENU_MARGIN, y);
    canvas.setTextColor((optionsmenu_cursor == i) ? COLOR_HIGHLIGHT : COLOR_NORMAL);
    canvas.print(options[i]->label);
    canvas.drawBitmap(OPTIONSMENU_PADDING + OPTIONSMENU_WIDTH - OPTIONSMENU_MARGIN - 8, y - 8, bmp_checkbox[options[i]->enabled], 8, 8, (optionsmenu_cursor == i) ? (options[i]->enabled ? COLOR_INDICATOR_OK : COLOR_INDICATOR_NG) : COLOR_NORMAL);
    y += OPTIONSMENU_LINEHEIGHT + OPTIONSMENU_LINESPACE;
  }

  if(justPressed(AWEXP_BUTTON_UP)) //press UP to move cursor up
  {
    optionsmenu_cursor--;
    if(optionsmenu_cursor < 0) { optionsmenu_cursor = NUM_OPTIONS - 1; }
  }
	else if(justPressed(AWEXP_BUTTON_DOWN)) //press DOWN to move cursor down
  {
    optionsmenu_cursor++;
    if(optionsmenu_cursor >= NUM_OPTIONS) { optionsmenu_cursor = 0; }
  }
  else if(justPressed(AWEXP_BUTTON_LEFT) || justPressed(AWEXP_BUTTON_RIGHT)) //press LEFT/RIGHT to toggle currently selected option on or off
  {
    options[optionsmenu_cursor]->enabled = !options[optionsmenu_cursor]->enabled;
  }
	else if(justPressed(AWEXP_BUTTON_SEL) || justPressed(AWEXP_BUTTON_OK) || justPressed(SHUTTER_BUTTON)) //press any other button to close options menu and save any changes
	{
    saveOptions();
    optionsmenu_open = 0;
	}
}