void setupDisplay()
{
	pinMode(TFT_BACKLIGHT, OUTPUT);
  digitalWrite(TFT_BACKLIGHT, 0);
	tft.init(240, 240);
	tft.setRotation(1);
  tft.drawRGBBitmap(0, 0, bmp_startup, 240, 240);
	digitalWrite(TFT_BACKLIGHT, 1);
  canvas.fillScreen(ST77XX_BLACK);
  canvas.setTextWrap(false);
  canvas.setFont(&PlanetMegapolis8px);
  clearStatus();
}

//transfer framebuffer to display
void updateDisplay()
{
  tft.drawRGBBitmap(0, 0, canvas.getBuffer(), 240, 240);
}

const uint8_t* status_bmp = NULL;
const char* status_message = NULL;
uint16_t status_bmp_color, status_text_color;

void drawViewfinderUI()
{
  //display image from camera on viewfinder
  memset(histogram, 0, sizeof(histogram));
  uint32_t focus_magnifier_x = (camera_frame->width - VIEWFINDER_WIDTH) / 2;
  uint32_t focus_magnifier_y = (camera_frame->height - VIEWFINDER_HEIGHT) / 2;
  byte focus_magnifier_active = millis() < focus_magnifier_end_time;
  float scale = float(camera_frame->width) / VIEWFINDER_WIDTH;
  for(uint32_t x = 0; x < VIEWFINDER_WIDTH; x++)
  {
    for(uint32_t y = 0; y < VIEWFINDER_HEIGHT; y++)
    {
      uint32_t addr;
      if(focus_magnifier_active)
      {
        addr = ((x + focus_magnifier_x) + ((y + focus_magnifier_y) * camera_frame->width));
      }
      else
      {
        addr = (uint32_t(x * scale) + uint32_t(y * scale) * camera_frame->width);
      }
      uint16_t color = getFramebufferColor(addr);
      canvas.getBuffer()[x + y * VIEWFINDER_WIDTH] = color;
      histogram[((color >> 12) + ((color >> 7) & 0b1111) + ((color >> 1) & 0b1111)) / 3]++; //calculate histogram. for each pixel's brightness, take average of red, green, and blue
    }
  }

  //draw parameter values
  for(byte i = 0; i < NUM_PARAMS; i++)
  {
    drawParam(i);
  }

  //draw indicators
  drawNumber(202, 182, photo_number, 4, ST77XX_WHITE); //draw number of photos taken on SD card
  canvas.drawBitmap(179, 185, bmp_shutteraf, 20, 8, (option_shutter_af->enabled) ? COLOR_INDICATOR_OK : ST77XX_BLACK);
  drawBattery();
  drawHistogram();
  if(option_thirds_grid->enabled) { drawThirdsGrid(); }
  if(option_horizon->enabled) { drawHorizon(); }
  drawStatus();
}

//draw a parameter value (bitmap icon & numeric value)
#define ICON_SIZE 18
#define ICON_PADDING 8
#define VALUE_PADDING 6
void drawParam(byte param)
{
  uint16_t color = (param == cursor_pos) ? COLOR_HIGHLIGHT : COLOR_NORMAL;
  byte x = param * (ICON_SIZE + ICON_PADDING);
  byte y = 240 - ICON_SIZE;
  canvas.drawBitmap(x, y, params[param].bitmap, ICON_SIZE, ICON_SIZE, color);
  y -= VALUE_PADDING + 14;
  if(isInAuto(param))
  {
    canvas.drawBitmap(x, y, param == PARAM_FOCUS ? bmp_text_af : bmp_text_au, 18, 14, (param != PARAM_FOCUS || blink_register) ? color : ST77XX_BLACK, ST77XX_BLACK);
  }
  else
  {
    int16_t value = params[param].ui_value;
    if(param == PARAM_WHITEBALANCE)
    {
      canvas.drawBitmap(x, y, bmp_wb_modes[value-1], 18, 14, color, ST77XX_BLACK);
    }
    else
    {
      drawNumber(x + ((params[param].digits == 1) * 10), y, value, params[param].digits, color);
      if(params[param].digits == 1)
      {
        canvas.drawBitmap(x, y + 2, bmp_up_arrow, 7, 4, value >= 0 ? color : ST77XX_BLACK);
        canvas.drawBitmap(x, y + 8, bmp_down_arrow, 7, 4, value <= 0 ? color : ST77XX_BLACK);
      }
    }
  }
}

//draw a number using my custom 7-segment font
void drawNumber(byte x, byte y, int16_t value, byte num_digits, uint16_t color)
{
  value = abs(value);
  for(int8_t i = num_digits - 1; i >= 0; i--)
  {
    canvas.drawBitmap(x + i * 10, y, bmp_digits[value % 10], 8, 14, color, ST77XX_BLACK);
    value /= 10;
  }
}

//draw battery indicator
#define NUM_VOLTAGES 25
#define VOLTAGE_DEADBAND 0.1
int8_t voltage_pos = 0, battery_level = -1;
float voltage_accum = 0, voltage = 0;
void drawBattery()
{
  //sample the voltage a bunch of times, then calculate the average of all the samples to smooth out noise
  voltage_accum += analogRead(BATT_MONITOR) * 2.0 * 3.3 / 4096;
  voltage_pos++;
  if(voltage_pos >= NUM_VOLTAGES)
  {
    voltage_accum /= NUM_VOLTAGES;
    //if the voltage has changed enough since the last reading, use it as the new battery level
    if(abs(voltage_accum - voltage) >= VOLTAGE_DEADBAND)
    {
      voltage = voltage_accum;
      if(voltage > 3.7) { battery_level = 3; }
      else if(voltage > 3.45) { battery_level = 2; }
      else if(voltage > 3.2) { battery_level = 1; }
      else { battery_level = 0; }
    }
    voltage_pos = 0;
    voltage_accum = 0;
  }
  //display battery level indicator once battery level has been determined
  if(battery_level > -1)
  {
    canvas.drawBitmap(225, 199, bmp_battery[battery_level], 15, 6, battery_level > 0 ? COLOR_INDICATOR_OK : COLOR_INDICATOR_NG, ST77XX_BLACK);
  }
}

//draw histogram
#define HISTOGRAM_SIZE 32
#define HISTOGRAM_X (240-HISTOGRAM_SIZE)
#define HISTOGRAM_Y (240-HISTOGRAM_SIZE)
void drawHistogram()
{
  canvas.fillRect(HISTOGRAM_X, HISTOGRAM_Y, HISTOGRAM_SIZE, HISTOGRAM_SIZE, COLOR_HISTOGRAM_BG);
	for(byte i = 0; i < 16; i++)
	{
		byte value = constrain(histogram[i] >> 8, 0, HISTOGRAM_SIZE);
		canvas.drawFastVLine(HISTOGRAM_X + i*2, HISTOGRAM_Y + HISTOGRAM_SIZE - value, value, COLOR_HISTOGRAM_FG);
    canvas.drawFastVLine(HISTOGRAM_X + i*2 + 1, HISTOGRAM_Y + HISTOGRAM_SIZE - value, value, COLOR_HISTOGRAM_FG);
	}
}

//draw rule-of-thirds grid
void drawThirdsGrid()
{
  for(byte i = 1; i < 3; i++)
  {
    canvas.drawFastHLine(0, map(i, 0, 3, 0, VIEWFINDER_HEIGHT), VIEWFINDER_WIDTH, COLOR_THIRDS_GRID);
    canvas.drawFastVLine(map(i, 0, 3, 0, VIEWFINDER_WIDTH), 0, VIEWFINDER_HEIGHT, COLOR_THIRDS_GRID);
  }
}

//draw artifical horizon using accelerometer
#define HORIZON_X (VIEWFINDER_WIDTH/2)
#define HORIZON_Y (VIEWFINDER_HEIGHT/2)
#define HORIZON_LINE_LENGTH 22
#define HORIZON_REF_LENGTH 12
void drawHorizon()
{
  float roll_angle = accelReadG(0) * 90; //x axis
  float pitch_angle = accelReadG(2) * 90; //z axis
  if(abs(roll_angle) < 30 && abs(pitch_angle) < 70)
  {
    float roll_radians = roll_angle * PI / 180;
    int16_t end_x = cos(roll_radians) * HORIZON_LINE_LENGTH;
    int16_t end_y = sin(roll_radians) * HORIZON_LINE_LENGTH;
    uint16_t color = (abs(roll_angle) < 2) ? COLOR_INDICATOR_OK : COLOR_THIRDS_GRID;
    canvas.drawLine(HORIZON_X - end_x, HORIZON_Y - end_y, HORIZON_X + end_x, HORIZON_Y + end_y, color);

    //draw flat reference lines
    canvas.drawFastHLine(HORIZON_X - HORIZON_LINE_LENGTH - HORIZON_REF_LENGTH, HORIZON_Y, HORIZON_REF_LENGTH, color);
    canvas.drawFastHLine(HORIZON_X + HORIZON_LINE_LENGTH + 1, HORIZON_Y, HORIZON_REF_LENGTH, color);
  }
}

//set big status icon and message
void showStatus(const uint8_t* bmp, char* msg, int16_t bmp_color, int16_t text_color)
{
  status_bmp = bmp;
  status_message = msg;
  status_bmp_color = bmp_color;
  status_text_color = text_color;
}
void showStatus(const uint8_t* bmp, char* msg, int16_t bmp_color)
{
  status_bmp = bmp;
  status_message = msg;
  status_bmp_color = bmp_color;
  status_text_color = ST77XX_WHITE;
}
//set status message only with no icon
void showStatus(char* msg)
{
  status_bmp = NULL;
  status_message = msg;
  status_text_color = ST77XX_WHITE;
}
//clear status messsage and icon
void clearStatus()
{
  status_bmp = NULL;
  status_message = "ready.";
  status_text_color = COLOR_NORMAL;
}

//draw big status message and icon if any are present
void drawStatus()
{
  if(status_bmp)
  {
    canvas.fillRoundRect(100, 70, 40, 40, 4, ST77XX_BLACK);
    canvas.drawBitmap(104, 74, status_bmp, 32, 32, status_bmp_color);
  }
  canvas.fillRect(0, 180, 178, 22, ST77XX_BLACK);
  if(status_message)
  {
    canvas.setCursor(0, 195);
    canvas.setTextColor(status_text_color);
    canvas.print(status_message);
  }
}