// Blossom, firmware for Adafruit Memento
// by Frost Sheridan
// https://www.cyberdragon.digital/dev/blossom/

#include <SPI.h>
#include <Wire.h>
#include "esp_camera.h"          //OV5640 camera sensor library, included with ESP32 arduino core
#include <Adafruit_AW9523.h>     //adafruit GPIO expander chip
#include <Adafruit_NeoPixel.h>   //adafruit neopixel
#include <Adafruit_GFX.h>        //adafruit graphics library
#include <Adafruit_ST7789.h>     //adafruit color LCD library
#include "PlanetMegapolis8px.h"
#include <SdFat.h>               //SD card file access driver
#include <Preferences.h>         //ESP32 non-volaitle preferences storage, included with ESP32 arduino core
#include "autofocus.h"
#include "bitmaps.h"
#include "startup_image.h"
#include <PNGenc.h>              //https://github.com/bitbank2/PNGenc


//camera
sensor_t *camera;
camera_fb_t *camera_frame = NULL;
camera_config_t camera_config;
int16_t focus_position = 0;
#define FOCUS_MAGNIFIER_TIME 2000 //focus magnifier stays on screen for this length of time
uint32_t focus_magnifier_end_time = 0;
byte old_focus_status = 0;
byte shutter_af_waiting = 0; //turns 1 to indicate that a photo should be taken once the current autofocus operation is complete (when shutter AF is enabled)

//display
Adafruit_ST7789 tft = Adafruit_ST7789(TFT_CS, TFT_DC, TFT_RESET);
GFXcanvas16 canvas(240, 240);
#define VIEWFINDER_WIDTH 240
#define VIEWFINDER_HEIGHT 180
uint16_t histogram[16];
int8_t cursor_pos = 0;
byte blink_register = 0;
#define COLOR_MESSAGE      0b1100011000011000
#define COLOR_ERROR        0b1111100000000000
#define COLOR_HIGHLIGHT    0b1111110100010100
#define COLOR_NORMAL       0b0100001000001000
#define COLOR_HISTOGRAM_BG 0b0010000100000100
#define COLOR_HISTOGRAM_FG 0b1111111111111111
#define COLOR_INDICATOR_OK 0b0000011111110000
#define COLOR_INDICATOR_NG 0b1111101100000000
#define COLOR_THIRDS_GRID  0b0001000010000010

//sd card
SdFat sd;
byte sdcard_ready = 0;
char photo_filename[13];
uint16_t photo_number = 0;
PNG png;

//preferences storage
Preferences preferences;
#define PARAM_SAVE_DELAY 3000 //wait this long after parameters have been modified to save them
uint32_t param_save_time = 0;

//buttons and other hardware
Adafruit_AW9523 aw;
Adafruit_NeoPixel neopixel(1, PIN_NEOPIXEL, NEO_GRB + NEO_KHZ800);
uint32_t last_button_state = 0xFFFFFFFF, button_state = 0xFFFFFFFF;
#define TONE_ERROR     130,200 //played when an error is occurred at startup
#define TONE_SHUTTERAF 659,10   //played when shutter AF is enabled and the shutter button was pressed
#define TONE_CAPTURE   1318,10  //played when a photo has been captured
#define TONE_SAVED     1760,10  //played when the captured photo has been saved to the SD card


//setup routine - load saved parameters and options, and initialize GPIO, display, camera, autofocus program, and SD card
void setup()
{
  Serial.begin(115200);
  setNeopixel(50, 20, 40);
  pinMode(SPEAKER, OUTPUT);
	digitalWrite(SPEAKER, 0);
	pinMode(SHUTTER_BUTTON, INPUT_PULLUP);
  setupParams();
  loadOptions();
  setupDisplay();
	if(!setupIOExpander()) { showFatalError("io expander fault"); }
  if(!setupAccel()) { showFatalError("accelerometer fault"); }
  if(!setupCamera()) { showFatalError("camera sensor fault"); }
  if(!setupAutofocus()) { showFatalError("autofocus fault"); }
  sdcard_ready = setupSDCard();
  loadParams();
  setNeopixel(0, 0, 0);
}


//halt everything, there's an error
void showFatalError(char* msg)
{
  setNeopixel(255, 0, 0);
  beep(TONE_ERROR);
  canvas.fillScreen(ST77XX_BLACK);
  canvas.drawBitmap(104, 104, bmp_fatal_error, 32, 32, ST77XX_WHITE);
  canvas.setCursor(0, 230);
  canvas.setTextColor(ST77XX_WHITE);
  canvas.print(msg);
  updateDisplay();

  while(1) {}
}